"""This file contains the logic defining all the parameters needed to \
set up a project with mloq."""
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import click
from omegaconf import DictConfig, MISSING, OmegaConf
import param

from mloq.config.configuration import as_resolved_dict, Configurable
from mloq.config.custom_click import confirm, prompt
from mloq.failure import MissingConfigValue


Choices = Union[List[str], Tuple[str], Set[str]]


class PromptParam:
    """
    Defines a configuration parameter.

    It allows to parse a configuration value from different sources in the following order:
        1. Environment variable named as MLOQ_PARAM_NAME
        2. Values defined in mloq.yaml
        3. Interactive promp from CLI (Optional)
    """

    def __init__(self, name: str, target: Configurable, **kwargs):
        """
        Initialize a ConfigParam.

        Args:
            name: Name of the parameter (as defined in mloq.yaml).
            text: Text that will be prompted in the CLI when using interactive mode.
            **kwargs: Passed to click.prompt when running in interactive mode.
        """
        self.name = name
        self._target = target
        prompt_text = self.param.doc if self.param.doc else self.name
        self._prompt_text = click.style(f"> {prompt_text}", fg="bright_magenta", reset=False)
        self._prompt_kwargs = kwargs
        self._prompt_kwargs["show_default"] = kwargs.get("show_default", True)
        self._prompt_kwargs["type"] = kwargs.get("type", str)

    @property
    def param(self) -> param.Parameter:
        """Get the param.Parameter object corresponding to the current configuration parameter."""
        return getattr(self._target.param, self.name)

    @property
    def value(self) -> Any:
        """Return the value of the configuration parameter."""
        return getattr(self._target, self.name)

    @property
    def config(self) -> Any:
        """Return the value of the parameter as defined in its config DictConfig."""
        if self.name not in self._target.config:
            raise MissingConfigValue(f"Config value {self.name} is not defined in config")
        elif OmegaConf.is_missing(self._target.config, self.name):
            return MISSING
        return self._target.config[self.name]

    def __call__(
        self,
        interactive: bool = False,
        default: Optional[Any] = None,
        **kwargs,
    ):
        """
        Return the value of the parameter parsing it from the different input sources available.

        Args:
            interactive: Prompt the user to input the value from CLI if it's not defined
                        in config or as en environment variable.
            default: Default value displayed in the interactive mode.
            **kwargs: Passed to click.prompt in interactive mode. Overrides the
                      values defined in __init__

        Returns:
            Value of the parameter.
        """
        value = default if default is not None else self.value
        value = self._prompt(value, **kwargs)
        return value

    def _prompt(self, value, **kwargs):
        """Prompt user for value."""
        _kwargs = dict(self._prompt_kwargs)
        _kwargs.update(kwargs)
        if value is not None:
            _kwargs["default"] = value
        return prompt(self._prompt_text, **_kwargs)


class MultiChoicePrompt(PromptParam):
    """
    Define a configuration parameter that can take multiple values \
    from a pre-defined set of values.

    It allows to parse a configuration value from different sources in the following order:
        1. Environment variable named as MLOQ_PARAM_NAME
        2. Values defined in mloq.yaml
        3. Interactive promp from CLI (Optional)
    """

    def __init__(
        self,
        name: str,
        target: Configurable,
        choices: Optional[Choices] = None,
        **kwargs,
    ):
        """
        Initialize a ConfigParam.

        Args:
            name: Name of the parameter (as defined in mloq.yaml).
            choices: Contains all the available values for the parameter.
            text: Text that will be prompted in the CLI when using interactive mode.
            **kwargs: Passed to click.prompt when running in interactive mode.
        """
        kwargs["type"] = str
        super(MultiChoicePrompt, self).__init__(name=name, target=target, **kwargs)
        self.choices = choices  # TODO: use this to validate user input.

    def _prompt(self, value, **kwargs) -> List[str]:
        """Transform the parsed string from the CLI into a list of selected values."""
        val = super(MultiChoicePrompt, self)._prompt(value, **kwargs)
        return self._parse_string(val) if isinstance(val, str) else val

    @staticmethod
    def _parse_string(value) -> List[str]:
        def filter_str(s):
            return s.lstrip().replace("'", "").replace('"', "").replace("[", "").replace("]", "")

        return [filter_str(s) for s in value.split(",")]


class StringPrompt(PromptParam):
    """
    Define a configuration parameter that can take a string value.

    It allows to parse a configuration value from different sources in the following order:
        1. Environment variable named as MLOQ_PARAM_NAME
        2. Values defined in mloq.yaml
        3. Interactive promp from CLI (Optional)
    """

    def __init__(
        self,
        name: str,
        target: Configurable,
        **kwargs,
    ):
        """
        Initialize a ConfigParam.

        Args:
            name: Name of the parameter (as defined in mloq.yaml).
            choices: Contains all the available values for the parameter.
            text: Text that will be prompted in the CLI when using interactive mode.
            **kwargs: Passed to click.prompt when running in interactive mode.
        """
        kwargs["type"] = str
        super(StringPrompt, self).__init__(name=name, target=target, **kwargs)


class IntPrompt(PromptParam):
    """
    Define a configuration parameter that can take an integer value.

    It allows to parse a configuration value from different sources in the following order:
        1. Environment variable named as MLOQ_PARAM_NAME
        2. Values defined in mloq.yaml
        3. Interactive promp from CLI (Optional)
    """

    def __init__(
        self,
        name: str,
        target: Configurable,
        **kwargs,
    ):
        """
        Initialize a ConfigParam.

        Args:
            name: Name of the parameter (as defined in mloq.yaml).
            choices: Contains all the available values for the parameter.
            text: Text that will be prompted in the CLI when using interactive mode.
            **kwargs: Passed to click.prompt when running in interactive mode.
        """
        kwargs["type"] = int
        super(IntPrompt, self).__init__(name=name, target=target, **kwargs)


class FloatPrompt(PromptParam):
    """
    Define a configuration parameter that can take a floating point value.

    It allows to parse a configuration value from different sources in the following order:
        1. Environment variable named as MLOQ_PARAM_NAME
        2. Values defined in mloq.yaml
        3. Interactive promp from CLI (Optional)
    """

    def __init__(
        self,
        name: str,
        target: Configurable,
        **kwargs,
    ):
        """
        Initialize a ConfigParam.

        Args:
            name: Name of the parameter (as defined in mloq.yaml).
            choices: Contains all the available values for the parameter.
            text: Text that will be prompted in the CLI when using interactive mode.
            **kwargs: Passed to click.prompt when running in interactive mode.
        """
        kwargs["type"] = float
        super(FloatPrompt, self).__init__(name=name, target=target, **kwargs)


class BooleanPrompt(PromptParam):
    """
    Defines a boolean configuration parameter.

    It allows to parse a configuration value from different sources in the following order:
        1. Environment variable named as MLOQ_PARAM_NAME
        2. Values defined in mloq.yaml
        3. Interactive promp from CLI (Optional)
    """

    def _prompt(self, value, **kwargs):
        """Prompt user for value."""
        _kwargs = dict(self._prompt_kwargs)
        _kwargs.update(kwargs)
        if "type" in _kwargs:
            del _kwargs["type"]
        if value is not None:
            _kwargs["default"] = value
        return confirm(self._prompt_text, **_kwargs)


PARAM_TO_PROMPT = {
    param.Boolean: BooleanPrompt,
    param.Integer: IntPrompt,
    param.Number: FloatPrompt,
    param.String: StringPrompt,
    param.ListSelector: MultiChoicePrompt,
    # TODO: MultiChoice, Choice
}


class Prompt:
    """
    Manage all the functionality needed to display a cli prompt.

    It allows to interactively define the values of the different parameters of a class.
    """

    def __init__(self, target: "Promptable"):
        """Initialize a Prompt."""
        self._target = target
        self._prompts = {}
        self._init_prompts()

    def __call__(self, key: str, inplace: bool = False, **kwargs) -> Any:
        """Display the a prompt to interactively define the parameter values of target."""
        return self.prompt(key=key, inplace=inplace, **kwargs)

    def _init_prompts(self) -> None:
        """Initialize the prompts corresponding to the target Promptable parameters."""
        self._prompts = {}
        conf: DictConfig = self._target.config
        param_ = self._target.param
        for name, value in as_resolved_dict(conf).items():
            param_inst = getattr(param_, name)
            type_ = type(param_inst)
            prompt_cls = PARAM_TO_PROMPT.get(type_)
            if prompt_cls is not None:
                default = value if value is not MISSING else param_inst.default
                self._prompts[name] = prompt_cls(name, self._target, default=default)

    def prompt(self, key: str, inplace: bool = False, **kwargs) -> Any:
        """Display the a prompt to interactively define the parameter values of target."""
        val = self._prompts[key](**kwargs)
        if inplace:
            setattr(self._target, key, val)
        else:
            return val

    def prompt_all(self, inplace: bool = False, **kwargs) -> Dict[str, Any]:
        """
        Prompt all the target's parameters.

        Return a dictionary containing the provided values.
        """

        def param_precedence(x):
            val = getattr(self._target.param, x).precedence
            return (1e100 if val is None else val), x

        sorted_keys = sorted(self._prompts.keys(), key=param_precedence)
        return {k: self.prompt(key=k, inplace=inplace, **kwargs) for k in sorted_keys}


class Promptable(Configurable):
    """
    Configurable class that allows to define the parameter values interactively using CLI prompts.

    It contains a prompt attribute in charge of managing the prompting functionality for the
    param.Parameters defined.
    """

    def __init__(self, **kwargs):
        """Initialize a Promptable."""
        super(Promptable, self).__init__(**kwargs)
        self.prompt = Prompt(self)



"""Mloq lint command implementation."""
from pathlib import Path

import click
from omegaconf import DictConfig

from mloq.command import Command
from mloq.config.param_patch import param
from mloq.files import ASSETS_PATH, file, pyproject_toml


lint_req = file(
    "requirements-lint.txt",
    ASSETS_PATH / "lint",
    "list of exact versions of the packages used to check your code style",
    is_static=True,
)
LINT_FILES = [pyproject_toml, lint_req]


class LintCMD(Command):
    """Implement the functionality of the lint Command."""

    cmd_name = "lint"
    files = tuple(LINT_FILES)
    disable = param.Boolean(default=None, doc="Disable lint command?")
    docstring_checks = param.Boolean(True, doc="Apply docstring checks?")
    pyproject_extra = param.String("", doc="Additional pyproject.toml configuration")
    project_name = param.String("${globals.project_name}", doc="Select project name")
    makefile = param.Boolean(True, doc="Add check and style commands to makefile")
    poetry_requirements = param.Boolean(True, doc="Add check and style commands to makefile")
    ignore_files = param.ListSelector(
        default=[],
        doc="Ignore the following files",
        objects=[f.dst for f in LINT_FILES],
    )

    def interactive_config(self) -> DictConfig:
        """Generate the configuration of the project interactively."""
        click.echo("Provide the values to generate the packaging files.")
        return self.parse_config()

    def record_files(self) -> None:
        """Register the files that will be generated by mloq."""
        for _file in self.files:
            if _file.dst not in self.record.config.lint.ignore_files:
                self.record.register_file(file=_file, path=Path())



"""Mloq license command implementation."""
from pathlib import Path

import click
from omegaconf import DictConfig

from mloq.command import Command
from mloq.config.param_patch import param
from mloq.files import ASSETS_PATH, file


TEMPLATES_PATH = ASSETS_PATH / "license"
dco = file(
    "DCO.md",
    TEMPLATES_PATH,
    "Developer Certificate of Origin - needed in open source projects to certify that "
    "the incoming contributions are legitimate",
    is_static=True,
)
mit_license = file("MIT_LICENSE", TEMPLATES_PATH, "license of the project", dst="LICENSE")
apache_license = file("APACHE_LICENSE", TEMPLATES_PATH, "license of the project", dst="LICENSE")
gpl_license = file(
    "GPL_LICENSE",
    TEMPLATES_PATH,
    "license of the project",
    dst="LICENSE",
    is_static=True,
)
LICENSES = {
    "MIT": mit_license,
    "Apache-2.0": apache_license,
    "GPL-3.0": gpl_license,
}

LICENSE_FILES = [file for file in LICENSES.values()] + [dco]


class LicenseCMD(Command):
    """Implement the functionality of the license Command."""

    cmd_name = "license"
    files = tuple(LICENSE_FILES)
    LICENSES = LICENSES
    disable = param.Boolean(default=None, doc="Disable license command?")
    license = param.String("MIT", doc="Project license type")
    copyright_year = param.Integer("${current_year:}", doc="Year when the project started")
    copyright_holder = param.String("${globals.owner}", doc="Copyright holder")
    project_name = param.String(doc="Select project name")
    project_url = param.String("${globals.project_url}", doc="GitHub project url")
    email = param.String("${globals.email}", doc="Owner contact email")

    def interactive_config(self) -> DictConfig:
        """Generate the configuration of the project interactively."""
        click.echo("Provide the values to generate the project license files.")
        return self.parse_config()

    def record_files(self) -> None:
        """Register the files that will be generated by mloq."""
        conf = self.record.config.license
        self.record.register_file(file=dco, path=Path())
        if conf.license != "proprietary":
            license_file = self.LICENSES[conf.license]
            self.record.register_file(file=license_file, path=Path())



"""The writer module defines the Writer class, which is in charge of creating the files \
and directories specified in the CMDRecord."""
import os
from pathlib import Path
from typing import Union

from omegaconf import DictConfig

from mloq.files import File, what_mloq_generated
from mloq.record import CMDRecord, Ledger
from mloq.templating import write_template


class Writer:
    """
    Write all the files specified on the provided CMDRecord.

    This class fills in rendered templates according to the provided
    configuration and generates the resulting file on the specified
    folder.

    Attributes of this class:
        path: Path string describing the destination folder where the
            file will be created.
        ledger: Instance of the Ledger class. It contains a dictionary
            summarizing the files that will by generated by mloq.
        record: Instance of the CMDRecord class. It keeps track of all
            files and directories that will be created from the
            user's configuration.
        overwrite: Boolean value. If True, existing files will be rewritten
            by mloq application.
    """

    def __init__(self, path: Union[Path, str], record: CMDRecord, overwrite: bool = False):
        """
        Initialize a new instance of the CMDRecord class.

        The class is instantiated from a CMDRecord object, which keeps
        of all files and directories that will be created by mloq.

        Args:
            path: Path string describing the location where the files
                will be generated.
            record: CMDRecord instance. Register the files and directories
                that will be created by mloq according to the configuration
                provided by the user.
            overwrite: Boolean value. If True, all existing files will be
                rewritten by the mloq application.
        """
        self._record = record
        self._ledger = Ledger()
        self._path = Path(path)  # Path to the project root directory
        self.overwrite = overwrite

    @property
    def path(self) -> Path:
        """Path string describing the location where the files will be generated."""
        return self._path

    @property
    def ledger(self) -> Ledger:
        """Keep track of the generated files."""
        return self._ledger

    @property
    def record(self) -> CMDRecord:
        """Register the files and directories generated by the mloq application."""
        return self._record

    def create_directories(self) -> None:
        """Create the folders registered inside the attribute 'record.directories'."""
        for directory in self.record.directories:
            os.makedirs(self.path / directory, exist_ok=True)

    def write_templates(self) -> None:
        """Generate the files recorded in the attribute 'record.files' on the specified path."""
        for path, file in self.record.files.items():
            self.write_template(file=file, path=path, config=self.record.config)

    def dump_ledger(self) -> None:
        """
        Write the summary of the generated files.

        This method collects the elements stored in ledger to create a markdown
        document that summarizes all the files generated by the MLOQ application.
        """
        config = DictConfig({**self.record.config, "generated_files": self.ledger.files})
        self.write_template(
            what_mloq_generated,
            path=what_mloq_generated.dst,
            config=config,
        )

    def write_template(
        self,
        file: File,
        path: Path,
        config: DictConfig,
    ) -> None:
        """
        Create new file containing the rendered template configuration.

        Args:
            file: File object representing the jinja template that will be rendered.
            path: Target folder where the generated files will be written.
            config: DictConfig containing the selected project configuration.

        Returns:
            None.
        """
        write_template(
            file=file,
            config=config,
            path=self.path / path,
            overwrite=self.overwrite,
            ledger=self.ledger,
        )

    def run(self) -> None:
        """Generate all files and directories registered inside the record instance."""
        self.create_directories()
        self.write_templates()
        self.dump_ledger()



"""This module contains the classes that keep track of the internal state of the\
 application when running a Command."""
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from omegaconf import DictConfig, OmegaConf

from mloq.files import File


class Ledger:
    """Keep track of the generated files."""

    def __init__(self):
        """Initialize a new instance of the Ledger class."""
        self._files = []

    @property
    def files(self) -> List[Tuple[str, str]]:
        """Return the list of generated file names."""
        return [(str(f), d) for f, d in sorted(self._files)]

    def register(self, file: Union[File, str, Path], description: Optional[str] = None) -> None:
        """Append another generated file to the book."""
        if isinstance(file, File):
            description = file.description if description is None else description
            file = file.dst
        elif description is None:
            raise ValueError("description is None. Please provide a file description")
        self._files.append((Path(file), description))


class CMDRecord:
    """
    Keep track of files and directories that will be created by mloq.

    The :class:`CMDRecord` acts as a single source of truth for storing the
    files and directories generated by `mloq` as well as the necessary configuration
    to generate them.

    This class registers three separate data sources:

    - `config`:
        An :class:`omegaconf.DictConfig` that contains the configuration of
        all the commands executed by `mloq`.

    - `files`:
        A `dict` containing the content and location of the files generated by mloq.
        It is indexed by :class:`Path` objects that indicate where the file will be created, and
        its values are instances of :class:`mloq.files.File`.

    - `directories`:
        List of :class:`Path` instances pointing to the different directories
        that `mloq` will create.

    This class is initialized from a configuration dictionary. The dictionary
    can be either an :class:`omegaconf.DictConfig` or an empty dictionary.
    """

    def __init__(
        self,
        config: Optional[DictConfig] = None,
        files: Optional[Dict[Path, File]] = None,
        directories: Optional[List[Path]] = None,
    ):
        """
        Initialize a new instance of :class:`CMDRecord`.

        Args:
           config: DictConfig element that stores the configuration parameters.
            files: Dictionary that registers the files generated by mloq. Keys
                are Paths strings labeling the location where the file will
                be created. Values are File elements referencing the current
                file.
            directories: List that stores the directories that will be generated
                by the mloq according to the user's configuration.
        """
        self._files = {} if files is None else files
        self._directories: List[Path] = [] if directories is None else directories
        self._config: DictConfig = DictConfig({}) if config is None else config

    @property
    def config(self) -> DictConfig:
        """
        Store the configuration parameters that govern the project's structure.

        It contains one configuration entry per each :class:`Command` that will be run,
        and each entry will only be modified by the :class:`Command` it represents. Each
        :class:`Command` instance is responsible for updating its corresponding
        `config` values.

        Returns:
             :class:`omegaconf.DictConfig` that contains the configuration of
             all the commands executed by `mloq`.
        """
        return self._config

    @property
    def files(self) -> Dict[Path, File]:
        """
        Return the dictionary of files used by mloq to generate the project configuration.

        `files` is a `dict` containing the content and location of the files generated by mloq.
        It is indexed by :class:`Path` objects that indicate where the file will be created, and
        its values are instances of :class:`mloq.files.File`.

        Each different :class:`Command` is responsible for registering the files it generates
        to the `files` dictionary.
        """
        return self._files

    @property
    def directories(self) -> List[Path]:
        """Contain the folders that will be created by mloq for storing the project's files."""
        return self._directories

    def update_config(self, config: DictConfig) -> None:
        """Update the configuration attribute according to the values entered by the user."""
        self._config = OmegaConf.merge(self._config, config)

    def register_file(
        self,
        file: File,
        path: Union[Path, str],
        description: Optional[str] = None,
    ) -> None:
        """
        Append a new file to the 'files' container.

        Keys are Path strings describing the location where the file will
        be created. Values are File objects containing the information
        of the file that will be generated.
        """
        if description is None and not file.description:
            raise ValueError("File description cannot be None. Please provide a description.")
        elif description is not None:
            file = File(
                name=file.name,
                src=file.src,
                dst=file.dst,
                description=description,
                is_static=file.is_static,
            )
        self.files[Path(path) / file.dst] = file

    def register_directory(self, path: Union[Path, str]) -> None:
        """Append a new directory path to the 'directories' container."""
        self.directories.append(Path(path))



"""Contains all the defined mloq Commands."""
from mloq.commands.ci import CiCMD
from mloq.commands.docker import DockerCMD
from mloq.commands.docs import DocsCMD
from mloq.commands.globals import GlobalsCMD
from mloq.commands.license import LicenseCMD
from mloq.commands.lint import LintCMD
from mloq.commands.package import PackageCMD
from mloq.commands.project import ProjectCMD
from mloq.commands.requirements import RequirementsCMD
from mloq.commands.setup import SetupCMD



"""Mloq setup command implementation."""
from pathlib import Path
from typing import List, Tuple

import click
from omegaconf import DictConfig

from mloq.command import Command, CommandMixin
from mloq.record import CMDRecord


def _sub_commands():
    from mloq.commands import (
        CiCMD,
        DockerCMD,
        DocsCMD,
        GlobalsCMD,
        LicenseCMD,
        LintCMD,
        PackageCMD,
        ProjectCMD,
        RequirementsCMD,
    )

    return (
        GlobalsCMD,
        ProjectCMD,
        PackageCMD,
        LicenseCMD,
        LintCMD,
        CiCMD,
        DocsCMD,
        DockerCMD,
        RequirementsCMD,
    )


SUB_COMMANDS = _sub_commands()


class SetupCMD(CommandMixin):
    """Implement the functionality of the setup Command."""

    cmd_name = "setup"
    files = tuple([file for cmd in SUB_COMMANDS for file in cmd.files])
    SUB_COMMAND_CLASSES = SUB_COMMANDS

    def __init__(self, record: CMDRecord, interactive: bool = False):
        """
        Initialize a SetupCMD class.

        Args:
            record: CMDRecord where the command data will be written.
            interactive: If True, parse the command configuration in interactive mode.
        """
        super(SetupCMD, self).__init__(record=record, interactive=interactive)
        self._sub_commands = [
            cmd(record=self.record, interactive=interactive) for cmd in self.SUB_COMMAND_CLASSES
        ]
        self.files = tuple([file for cmd in self._sub_commands for file in cmd.files])

    @property
    def config(self) -> DictConfig:
        """List of all the commands that will be executed when running mloq setup."""
        return self.record.config

    @property
    def sub_commands(self) -> List[Command]:
        """List of all the commands that will be executed when running mloq setup."""
        return self._sub_commands

    @property
    def directories(self) -> Tuple[Path]:
        """Tuple containing Paths objects representing the directories created by the command."""
        return tuple([directory for cmd in self.sub_commands for directory in cmd.directories])

    def interactive_config(self) -> DictConfig:
        """Generate the configuration of the project interactively."""
        click.echo("Provide the values to generate the project configuration.")
        for cmd in self.sub_commands:
            cmd.interactive_config()
        return self.record.config

    def parse_config(self) -> DictConfig:
        """Update the configuration DictConfig with the Command parameters."""
        for cmd in self.sub_commands:
            cmd.parse_config()
        return self.record.config

    def run_side_effects(self) -> None:
        """Apply additional configuration methods."""
        for cmd in self.sub_commands:
            cmd.run_side_effects()

    def record_files(self) -> None:
        """Register the files that will be generated by mloq."""
        for cmd in self.sub_commands:
            cmd.record_files()



"""Patch param to allow omegaconf Missing values and interpolation strings."""
from omegaconf import MISSING
import param as param_


__DEFAULT_MARK = "__DEFAULT_MARK__"


def __init__patched(self, default=__DEFAULT_MARK, **kwargs):
    """Handle missing values and interpolations trings when initializing a param.Parameter."""
    from mloq.config.configuration import is_interpolation

    is_missing = default == MISSING
    interpolated = is_interpolation(default)
    self._missing_init = is_missing
    self._interpolation_init = interpolated
    if is_missing or interpolated:
        kwargs["allow_None"] = True
        default = None
    elif default == __DEFAULT_MARK:
        param_class = getattr(param_, self.__class__.__name__)
        default = param_class().default
    super(self.__class__, self).__init__(default=default, **kwargs)


def _create_param__(item):
    """Patch the target param.Parameter to support missing values and interpolations strings."""
    base = getattr(param_, item)
    patched_class = type(
        item,
        (base,),
        {
            "__init__": __init__patched,
            "__slots__": list(base.__slots__) + ["_missing_init", "_interpolation_init"],
        },
    )
    return patched_class


PATCHED_PARAMETERS = {"String", "Integer", "Number", "Tuple", "Dict", "List", "ListSelector"}


class __ParamPatcher:
    """Patch the param package to handle missing values and interpolation strings."""

    PATCHED_PARAMETERS = PATCHED_PARAMETERS

    def __getattr__(self, item):
        """Patch the parameters included in PATCHED_PARAMETERS."""
        if item in PATCHED_PARAMETERS:
            return _create_param__(item)
        return getattr(param_, item)

    def __setattr__(self, key, value):
        """Read only monkeypatching."""
        raise NotImplementedError


param = __ParamPatcher()



"""
This module defines the Configurable class and associated logic.

The Configurable class extends the param.Parameterizable class to keep track of
the class parameters using an omegaconf.DictConfig.
"""
import copy
import dataclasses
from dataclasses import field, make_dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

import omegaconf
from omegaconf import Container, MISSING, OmegaConf
from omegaconf.errors import InterpolationToMissingValueError, MissingMandatoryValue
import param as param__
from typing_extensions import Protocol

from mloq.config.param_patch import param


#
try:
    from mypy.typeshed.stdlib.dataclasses import Field

    DClassField = Field[Any]
except ModuleNotFoundError:
    DClassField = Any


class Dataclass(Protocol):
    """Type hinting to defined a dataclass as a typing Protocol."""

    # as already noted in comments, checking for this attribute is currently
    # the most reliable way to ascertain that something is a dataclass
    __dataclass_fields__: Dict


ConfigValue = Any

ConfigurationDict = Union[Dict[Union[str, int, Enum, float, bool], ConfigValue]]

PythonType = Union[float, int, bool, str, list, dict, tuple]

ParamType = Union[
    param.Number,
    param.Integer,
    param.Boolean,
    param.String,
    param.Array,
    param.List,
    param.Dict,
    param.Tuple,
]

DataClassDict = Dict[str, Tuple[type, DClassField]]


class DictConfig(param.ClassSelector):
    """param.Parameter that defines a DictConfig object."""

    def __init__(
        self,
        default: Optional[omegaconf.DictConfig] = None,
        doc: Optional[str] = None,
        instantiate: bool = True,
        per_instance: bool = True,
        **kwargs,
    ):
        """
        Initialize a DictConfig.

        Args:
            default: Default value of the Parameter. Must be an instance of DictConfig.
            doc: Documentation of the Parameter.
            instantiate: Expect and instance of DictConfig.
            per_instance: Create a new DictConfig instance every time you initialize a
                         Parameterized class.
            **kwargs: Passed to param.ClassSelector.__init__.
        """
        default = omegaconf.DictConfig({}) if default is None else default
        if doc is None:
            doc = "Structured omegaconf.DictConfig representing the param.Parameters information"
        kwargs["class_"] = omegaconf.DictConfig

        super(DictConfig, self).__init__(
            default=default,
            doc=doc,
            instantiate=instantiate,
            per_instance=per_instance,
            **kwargs,
        )


PARAM_TO_TYPE = {
    param.Boolean: bool,
    param.Integer: int,
    param.Number: float,
    param.String: str,
    param.List: list,
    param.ListSelector: list,
    param.Dict: dict,
    param.Tuple: tuple,
    param__.Boolean: bool,
    param__.Integer: int,
    param__.Number: float,
    param__.String: str,
    param__.List: list,
    param__.ListSelector: list,
    param__.Dict: dict,
    param__.Tuple: tuple,
    DictConfig: omegaconf.DictConfig,
}


def param_to_dataclass_dict(
    obj: Union[param.Parameterized, Any],
) -> Dict[str, Tuple[type, DClassField]]:
    """
    Create a dictionary that can be used to initialize a dataclass containing the parameters \
    of the target param.Parameterized class.

    Args:
        obj: Class or instance of a param.Parameterized class.

    Returns:
        dict containing the fields required to define a dataclass with the obj parameters.
    """
    data = {}
    for k, v in obj.params().items():
        if k in ["name", "config"]:
            continue
        for param_type, _type in PARAM_TO_TYPE.items():
            if isinstance(v, param_type):
                value = getattr(obj, k) if isinstance(obj, param.Parameterized) else v.default
                data[k] = (_type, field(default=value))
                break
    return data


def param_to_dataclass(obj: Union[param.Parameterized, Any]) -> type:
    """Create a dataclass equivalent to the target param.Parameterized target."""
    name = obj.__class__.name if isinstance(obj, param.Parameterized) else obj.name
    # FIXME: Assumes all keys are strings
    datac = make_dataclass(
        name,
        [(str(k), t, v) for k, (t, v) in param_to_dataclass_dict(obj).items()],
    )
    return datac


def param_to_omegaconf(obj: Union[param.Parameterized, Any]) -> omegaconf.DictConfig:
    """Transform a param.Parameterized class into an OmegaConf structured configuration."""
    return OmegaConf.structured(param_to_dataclass(obj))


def is_interpolation(s: str) -> bool:
    """Return True if the provided string is an OmegaConf interpolation string."""
    if not isinstance(s, str):
        return False
    return "${" in s and "}" in s  # TODO: use regex


def to_param_type(obj: param.Parameterized, config: DictConfig, key: str) -> Any:
    """Transform the provided attribute of the target param.Parameterized object \
    into the appropriate type so it can be stored in a configuration file."""
    # Yaml cannot handle tuples, so we convert the value
    config = copy.deepcopy(config)
    OmegaConf.resolve(config)
    param_obj = obj.param.params().get(key)
    value = config[key] if not OmegaConf.is_missing(config, key) else param_obj.default
    if isinstance(value, omegaconf.ListConfig):
        value = [x for x in value]
    elif isinstance(value, omegaconf.DictConfig):
        value = {**value}
    type_ = PARAM_TO_TYPE.get(param_obj.__class__)
    if type_ and not isinstance(value, type_):
        if value is None:
            value = value if param_obj.allow_None else type_()
        else:
            value = type_(value)  # if value is not None else type_()

    return value


def to_config(
    config: Union[
        omegaconf.DictConfig,
        ConfigurationDict,
        Dataclass,
        param.Parameterized,
        None,
    ],
    **kwargs,
) -> omegaconf.DictConfig:
    """Transform the provided object into an omegaconf.DictConfig."""
    if isinstance(config, param.Parameterized):
        config = param_to_omegaconf(config)
    elif dataclasses.is_dataclass(config):
        config = OmegaConf.structured(config, **kwargs)
    elif not isinstance(config, omegaconf.DictConfig):
        config = OmegaConf.create(config, **kwargs)
    elif config is None:
        return omegaconf.DictConfig({})
    return config


def resolve_as_dict(
    obj,
    config: Union[omegaconf.DictConfig, ConfigurationDict, Dataclass, param.Parameterized],
    **kwargs,
) -> ConfigurationDict:
    """Transform the provided object into a dictionary resolving all its interpolations."""
    config: Union[Container, omegaconf.DictConfig] = to_config(config, **kwargs)
    OmegaConf.resolve(config)
    param_data: Dict[str, Any] = {k: to_param_type(obj, config, k) for k in config}
    return param_data


def safe_select(cfg: DictConfig, key: str, default: Any = None) -> Any:
    """
    Access safely the target value of the provided cfg DictConfig.

    Return  MISSING if the value cannot be resolved or it's missing.
    """
    try:
        return OmegaConf.select(
            cfg=cfg,
            key=key,
            default=default,
            throw_on_resolution_failure=True,
            throw_on_missing=True,
        )
    except (MissingMandatoryValue, InterpolationToMissingValueError):  # , InterpolationKeyError):
        return MISSING


def as_resolved_dict(cfg: DictConfig) -> ConfigurationDict:
    """Return a dictionary containing the resolved values for the provided DictConfig."""
    resolved_dict = {k: safe_select(cfg, k) for k in cfg.keys()}
    return resolved_dict


class OmegaConfInterface:
    """Common functionality to work with configurations."""

    def __init__(self, target: "Configurable", allow_missing: bool = False):
        """
        Initialize an OmegaConfInterface.

        Args:
            target: Keep track of target param values using a DictConfig.
            allow_missing: Allow missing values in the target configuration.
        """
        self._target = target
        self.allow_missing = allow_missing

    @property
    def config(self) -> omegaconf.DictConfig:
        """Return a DictConfig containing the target configuration."""
        return self._target.config

    @property
    def interpolations(self) -> ConfigurationDict:
        """Return a dictionary containing the interpolations of the target configuration."""
        cont = OmegaConf.to_container(self.config, resolve=False)
        return {k: v for k, v in cont.items() if OmegaConf.is_interpolation(self.config, str(k))}

    @property
    def missing(self) -> List[Union[str, int, Enum, float, bool]]:
        """Return a list containing the names of the configuration that are MISSING."""
        return [k for k, v in as_resolved_dict(self.config).items() if v == MISSING]

    def _resolve_inplace(self, key: Optional[str] = None) -> None:
        """Resolve and update the target attribute if it's an interpolation string."""
        if key is None:
            OmegaConf.resolve(self._target.config)
            return
        self.config[key] = self.select(key=key)

    def resolve(
        self,
        key: Optional[str] = None,
        inplace: bool = False,
    ) -> Union[Container, ConfigValue, None]:
        """
        Resolve the target attribute if it is an interpolation string.

        Args:
            key: Name of the target's attribute to be resolved.
            inplace: If True, update the configuration value replacing the
                    interpolation string with the resolved value.

        Returns:
            Resolved value of the target attribute.
        """
        if inplace:
            return self._resolve_inplace(key)
        value = as_resolved_dict(self.config) if key is None else self.select(key=key)
        return value

    def is_missing(self, key: str) -> bool:
        """Return True if the key target's attribute is Missing, otherwise return False."""
        return safe_select(self.config, key) == MISSING

    def is_interpolation(self, key: str) -> bool:
        """Check if the key target's attribute is an interpolation string."""
        return OmegaConf.is_interpolation(self.config, key)

    def select(self, key, default=None) -> Any:
        """
        Select the key target's attribute.

        Return MISSING if key corresponds to a missing value, or an
        interpolation that resolves to a missing value.
        """
        return safe_select(self.config, key=key, default=default)


class BaseConfig(OmegaConfInterface):
    """Manages getters and setters to access the target's configuration values."""

    def __init__(
        self,
        target: "Configurable",
        config: Optional[Union[ConfigurationDict, omegaconf.DictConfig]] = None,
        cfg_node: Optional[str] = None,
        allow_missing: bool = False,
        **kwargs,
    ):
        """Initialize a BaseConfig."""
        super(BaseConfig, self).__init__(target=target, allow_missing=allow_missing)
        self._setup_config(config, cfg_node=cfg_node, **kwargs)

    def __getitem__(self, item: str) -> Any:
        """Access the target config value."""
        return self.config[item]

    def __setitem__(self, key: str, value) -> Any:
        """Set the target config value."""
        self.config[key] = value

    def to_container(self, resolve: bool = False, **kwargs) -> Container:
        """Return a container containing the target's configuration."""
        try:
            return OmegaConf.to_container(self.config, resolve=resolve, **kwargs)
        except (MissingMandatoryValue, InterpolationToMissingValueError):
            d = omegaconf.DictConfig(as_resolved_dict(self.config))
            return OmegaConf.to_container(d, resolve=resolve, **kwargs)

    @staticmethod
    def _resolve_node(
        kwargs: dict,
        config: Optional[omegaconf.DictConfig] = None,
        cfg_node: Optional[str] = None,
    ) -> omegaconf.DictConfig:
        """Return a DictConfig containing the resolved configuration values defined in kwargs."""
        kwsconf = OmegaConf.create(kwargs)
        if not config:
            return kwsconf
        # FIXME: IF we resolve at init to get the global conf value we loose the interpolations
        is_node = config and cfg_node is not None
        resolved_node = config  # omegaconf.DictConfig(as_resolved_dict(config))
        if is_node and cfg_node in config:
            resolved_node = config[cfg_node]
        resolved_with_kws = OmegaConf.merge(kwsconf, resolved_node)

        if is_node:
            # node_conf = OmegaConf.create({cfg_node: resolved_with_kws})
            # full_conf = OmegaConf.merge(config, node_conf)
            # OmegaConf.resolve(full_conf)
            # full_conf = OmegaConf.create(as_resolved_dict(full_conf))
            config[cfg_node] = resolved_with_kws
            return config[cfg_node]
        return resolved_with_kws

    def _setup_config(
        self,
        config: Optional[Union[ConfigurationDict, omegaconf.DictConfig]] = None,
        cfg_node: Optional[str] = None,
        **kwargs,
    ):
        """Initialize and validate the structured config of target."""
        conf = self._resolve_node(kwargs=kwargs, cfg_node=cfg_node, config=config)
        OmegaConf.set_struct(conf, True)
        self._target.config = conf  # TODO: make param.config constant


class Config(BaseConfig):
    """
    Config handles the `.conf` attribute of a Configurable class.

    It is analogous to `.param` for param.Parameterized classes.
    This class implements all the logic to access and update the config attribute
    of a Configurable class, which returns a DictConfig instance that
    is automatically update when the parameters of the class change.
    """

    @property
    def params(self) -> Dict[str, param.Parameter]:
        """Return the param.Parameter dictionary of the target configurable."""
        return self._target.param.params()

    def resolve(
        self,
        key: Optional[str] = None,
        inplace: bool = False,
    ) -> Union[Container, ConfigValue, None]:
        """Resolve the key attribute of the target Configurable."""
        rsl = super(Config, self).resolve(inplace=inplace)
        if not inplace:
            value = rsl if key is None else self.to_param_type(key)
            return value

    def to_param_type(self, key) -> Any:
        """Transform the value of the key target's parameter to a DictConfig compatible type."""
        value = self.select(key)
        param_obj = self.params.get(key)
        if value == MISSING:
            return param_obj.default
        # Yaml cannot handle python data types such as tuples, so we cast value
        # to the appropriate type after reading data from the the DictConfig
        # and before setting the corresponding param.key
        if isinstance(value, omegaconf.ListConfig):
            value = [x for x in value]
        type_ = PARAM_TO_TYPE.get(param_obj.__class__)
        if isinstance(param_obj, param.Tuple):
            raise ValueError("TUPLE")
        if type_ and not isinstance(value, type_):
            if value is None:
                value = value if param_obj.allow_None else type_()
            else:
                value = type_(value)
        return value

    def dataclass_dict(
        self,
        ignore: Optional[Union[list, set, tuple, str]] = None,
    ) -> DataClassDict:
        """Return a dictionary to create a dataclass with the target's parameters."""
        data = {}
        ignored = {"name", "config"} if ignore is None else ignore
        ignored = set([ignored]) if isinstance(ignored, str) else ignored
        for k, v in self.params.items():
            if k in ignored:
                continue
            for param_type, _type in PARAM_TO_TYPE.items():
                if isinstance(v, param_type):
                    value = v.default if self._target is None else getattr(self._target, k)
                    data[k] = (_type, field(default=value))
                    break
        return data

    def to_dataclass(self) -> type:  # DataClass class, not an instance
        """Return a dataclass describing the parameter values of the target Configurable."""
        tgt = self._target
        name = tgt.__class__.__name__ if isinstance(tgt, Configurable) else tgt.__name__
        dclass = make_dataclass(name, [(k, t, v) for k, (t, v) in self.dataclass_dict().items()])
        return dclass

    def to_dictconfig(self) -> DictConfig:
        """Return a structured DictConfig containing the parameters of the target Configurable."""
        return OmegaConf.structured(self.to_dataclass())

    def sync(self):
        """Ensure the parameter values of the target class have the right type."""
        for k in self.config.keys():
            super(Configurable, self._target).__setattr__(k, self.to_param_type(k))

    def _setup_config(
        self,
        config: Optional[Union[ConfigurationDict, DictConfig]] = None,
        cfg_node: Optional[str] = None,
        **kwargs,
    ):
        """Initialize and validate the structured config of target."""
        ignored = {"name", "config"}
        # Make sure the DictConfig is initialized with all the params as keys
        kwargs = {k: kwargs.get(k, v.default) for k, v in self.params.items() if k not in ignored}
        super(Config, self)._setup_config(config=config, cfg_node=cfg_node, **kwargs)
        self.sync()


CONF_ATTRS = {"config", "conf", "_conf"}


class Configurable(param.Parameterized):
    """
    A Configurable class is an extension of param.Parameterized that allows to handle parameters \
    with missing values and omegaconf interpolation strings.

    It add a config attribute containing an omegaconf.DictConfig that contains the values of the
    class param.Parameters.

    It also provides a `conf` attribute that allows to access omegaconf functionality
    for managing configurations in a similar fashion as the `param` attribute allows to access
    param.Parameter functionality.
    """

    config = DictConfig(readonly=False, per_instance=True, instantiate=True)

    def __init__(
        self,
        config: Optional[Union[ConfigurationDict, DictConfig]] = None,
        throw_on_missing: bool = True,
        cfg_node: Optional[str] = None,
        **kwargs,
    ):
        """Initialize a Configurable."""
        interp_kwargs = resolve_as_dict(self, kwargs)
        super(Configurable, self).__init__(**interp_kwargs)
        self.__conf = Config(
            self,
            config,
            cfg_node=cfg_node,
            throw_on_missing=throw_on_missing,
            **kwargs,
        )

    @property
    def conf(self) -> Config:
        """Access the Config instance that tracks and manages the values in the class config."""
        return self.__conf

    def __setattr__(self, key, value):
        """Update the config values when setting a parameter."""
        is_interp = is_interpolation(value)
        if value == MISSING or is_interp:
            self.config[key] = value
            value = (
                self.conf.to_param_type(key=key) if is_interp else self.param.params()[key].default
            )
        # Update the config dict as well as the parameters. Ignored during __init__ of parent class
        elif key in self.param.params() and hasattr(self, "conf"):
            self.config[key] = value
            value = self.conf.to_param_type(key=key)

        super(Configurable, self).__setattr__(key, value)

    def __getattr__(self, item):
        """Add support for MISSING values when accessing the parameter values."""
        if item != "config" and OmegaConf.is_missing(self.config, item):
            return MISSING
        return super(Configurable, self).__getattr__(item)



"""This module defines the pipeline for running a mloq command, such as config loading, \
template writing and interfacing with click."""
from pathlib import Path
import sys
from typing import Callable, Union
from unittest.mock import patch

import hydra
from omegaconf import DictConfig, OmegaConf

from mloq import _logger
from mloq.command import Command
from mloq.files import mloq_yml
from mloq.record import CMDRecord
from mloq.writer import Writer


def load_config(config_file: Union[Path, str], hydra_args: str) -> DictConfig:
    """
    Load the necessary configuration for running mloq from a mloq.yaml file.

    If no path to mloq.yaml is provided, it returns a template to be filled in
    using the interactive mode.

    Args:
        config_file: Path to the target mloq.yaml file.
        hydra_args: Arguments passed to hydra for composing the project configuration.

    Returns:
        DictConfig containing the project configuration.
    """
    config_file = Path(config_file) if config_file else Path() / mloq_yml.dst
    config_file = (
        config_file
        if (config_file.exists() and config_file.is_file())
        else config_file / mloq_yml.dst
    )
    if config_file.exists() and config_file.is_file():
        _logger.info(f"Loading config file from {config_file}")
        hydra_args = ["--config-dir", str(config_file.parent)] + list(hydra_args)
        config = DictConfig({})

        @hydra.main(config_path=".", config_name=config_file.name)
        def load_config(loaded_config: DictConfig):
            nonlocal config
            config = loaded_config

        with patch("sys.argv", [sys.argv[0]] + list(hydra_args)):
            load_config()
    else:
        _logger.info("No mloq.yaml file provided. Creating a new configuration")
        config = OmegaConf.load(mloq_yml.src)
    return config


def write_record(
    record: CMDRecord,
    path: Union[Path, str],
    overwrite: bool = False,
    only_config: bool = False,
) -> None:
    """
    Write the contents of the provided record to the target path.

    The writing process is performed by :class: `Writer`, class that fills
    in rendered templates according to the given configuration.

    Args:
        record: CMDRecord containing all the data to be written.
        path: Target directory to write the data.
        overwrite: If True overwrite existing files.
        only_config: Do not write any file except mloq.yaml

    Returns:
        None.
    """
    if only_config:
        with open(Path(path) / mloq_yml.dst, "w") as f:
            OmegaConf.save(config=record.config, f=f)
    else:
        writer = Writer(record=record, overwrite=overwrite, path=path)
        writer.run()


def run_command(cmd_cls, use_click: bool = True) -> Callable:
    """
    Run the given Command class.

    Args:
        cmd_cls: Command to be executed.
        use_click: If True run the function as a "cli" command.

    Returns:
        A function that will run the target class as a mloq command.
    """
    from mloq.cli import mloq_click_command

    def _run_command(
        config_file: str,
        output_directory: str,
        overwrite: bool,
        only_config: bool,
        interactive: bool,
        hydra_args: str,
    ) -> None:
        config: DictConfig = load_config(config_file=config_file, hydra_args=hydra_args)
        record = CMDRecord(config=config)
        cmd: Command = cmd_cls(record=record, interactive=interactive)
        record = cmd.run()
        write_record(
            record=record,
            path=output_directory,
            overwrite=overwrite,
            only_config=only_config,
        )

    if use_click:
        _run_command = mloq_click_command(_run_command)
    _run_command.__name__ = cmd_cls.cmd_name
    return _run_command



"""Command line interface for mloq."""
import os
from pathlib import Path
from typing import Callable, Optional

import click

from mloq.runner import run_command
from mloq.version import __version__


overwrite_opt = click.option(
    "--overwrite/--no-overwrite",
    "-o/ ",
    default=False,
    show_default=True,
    help="Value indicating whether to overwrite existing files.",
)
config_file_opt = click.option(
    "--filename",
    "-f",
    "config_file",
    default=None,
    show_default=True,
    help="Name of the repository config file",
    type=click.Path(exists=True, file_okay=True, dir_okay=True, resolve_path=True),
)
only_config_opt = click.option(
    "--only-config/--everything",
    "-c/ ",
    default=False,
    show_default=True,
    help="Value indicating whether to not generate all the files except mloq.yaml.",
)
output_directory_arg = click.argument(
    "output_directory",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
)

interactive_opt = click.option(
    "--interactive/--no-interactive",
    "-i/ ",
    default=False,
    show_default=True,
    help="If True the configuration values will be defined interactively on the command line.",
)

hydra_args = click.argument("hydra_args", nargs=-1, type=click.UNPROCESSED)


def mloq_click_command(func):
    """Wrap a command function to interface with click."""
    func = hydra_args(func)
    func = only_config_opt(func)
    func = interactive_opt(func)
    func = overwrite_opt(func)
    func = output_directory_arg(func)
    func = config_file_opt(func)
    func = click.command(context_settings=dict(ignore_unknown_options=True))(func)
    return func


class MloqCLI(click.MultiCommand):
    """Load the commands available at runtime from the files present in the command module."""

    command_folder = Path(__file__).parent / "commands"

    def list_commands(self, ctx):
        """List the names of the mloq commands available."""
        rv = []
        for filename in os.listdir(self.command_folder):
            if filename.endswith(".py") and filename != "__init__.py":
                rv.append(filename[:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name) -> Callable:
        """Create the command callable corresponding to the provided command name."""
        ns = {}
        fn = os.path.join(self.command_folder, name + ".py")
        with open(fn) as f:
            code = compile(f.read(), fn, "exec")
            eval(code, ns, ns)
        command_class = ns[f"{name.capitalize()}CMD"]
        # TODO: handle exit codes if needed
        return run_command(command_class)


@click.command(cls=MloqCLI)
def cli():  # noqa: D103
    pass


def welcome_message(extra: bool = False, string: Optional[str] = None):
    """Welcome message to be displayed during interactive setup."""
    click.echo(f"Welcome to the MLOQ {__version__} interactive setup utility.")
    if extra:
        click.echo(f"{string}")
    click.echo()
    click.echo(
        "Please enter values for the following settings (just press Enter "
        "to accept a default value, if one is given in brackets).",
    )
    click.echo()



"""Mloq globals command implementation."""
import click
from omegaconf import DictConfig, OmegaConf
import param

from mloq.command import Command


class GlobalsCMD(Command):
    """Implement the functionality of the globals Command."""

    cmd_name = "globals"
    project_name = param.String(doc="Select project name")
    description = param.String(doc="Short description of the project")
    author = param.String(doc="Author(s) of the project")
    owner = param.String("${globals.author}", doc="Github handle of the project owner")
    email = param.String(doc="Owner contact email")
    open_source = param.Boolean(doc="Is the project Open Source?")
    project_url = param.String("???", doc="GitHub project url")
    default_branch = param.String(doc="Default branch of the project")
    license = param.String("MIT", doc="Project license type")
    use_poetry = param.Boolean(True, doc="Use poetry to manage dependencies?")

    def interactive_config(self) -> DictConfig:
        """Generate the configuration of the project interactively."""
        click.echo("The following values will occur in several places in the generated files.")
        return super(GlobalsCMD, self).interactive_config()

    def parse_config(self) -> DictConfig:
        """Generate the configuration of the project via a configuration file."""
        default_url = f"https://github.com/{self.owner}/{self.project_name.replace(' ', '-')}"
        self.project_url = OmegaConf.select(self.config, "project_url", default=default_url)
        self.conf.sync()
        return super(GlobalsCMD, self).parse_config()



"""Mloq docs command implementation."""
from pathlib import Path
from typing import Tuple

import click
from omegaconf import DictConfig
import param

from mloq.command import Command
from mloq.files import ASSETS_PATH, file


DOCS_ASSETS_PATH = ASSETS_PATH / "docs"
# Documentation files
makefile_docs = file(
    "makefile_docs.txt",
    DOCS_ASSETS_PATH,
    "common make commands for building the documentation",
    dst="Makefile",
    is_static=True,
)
make_bat_docs = file(
    "make_bat.txt",
    DOCS_ASSETS_PATH,
    "common make commands for building the documentation",
    dst="make.bat",
    is_static=True,
)
docs_req = file(
    "requirements-docs.txt",
    DOCS_ASSETS_PATH,
    "list of exact versions of the packages needed to build your documentation",
    is_static=True,
)
conf_py = file(
    "conf.txt",
    DOCS_ASSETS_PATH,
    "configuration file for sphinx and doc plugins",
    dst="conf.py",
)
index_md = file(
    "index.md",
    DOCS_ASSETS_PATH,
    "configuration file for sphinx and doc plugins",
)
deploy_docs = file(
    "deploy-docs.yml",
    DOCS_ASSETS_PATH,
    "workflow for deploying the documentation to GitHub pages",
)

DOCS_FILES = [conf_py, index_md, makefile_docs, make_bat_docs, docs_req, deploy_docs]


class DocsCMD(Command):
    """Implement the functionality of the docs Command."""

    cmd_name = "docs"
    disable = param.Boolean(False, doc="Disable docs command?")
    project_name = param.String(doc="Select project name")
    description = param.String(doc="Short description of the project")
    author = param.String("${globals.author}", doc="Author(s) of the project")
    copyright_year = param.Integer(doc="Year when the project started")
    copyright_holder = param.String("${docs.author}", doc="Copyright holder")
    deploy_docs = param.Boolean(True, doc="Deploy docs to GitHub Pages?")
    default_branch = param.String("${globals.default_branch}", doc="Branch used to build the docs")
    project_url = param.String("${globals.project_url}", doc="GitHub project url")
    files = tuple(DOCS_FILES)

    @property
    def directories(self) -> Tuple[Path]:
        """Tuple containing Paths objects representing the directories created by the command."""
        docs_folder = Path("docs") / "source" / "markdown"
        paths = [docs_folder]
        if self.deploy_docs:
            workflow_path = Path(".github") / "workflows"
            paths.append(workflow_path)
        return tuple(paths)

    def interactive_config(self) -> DictConfig:
        """Generate the configuration of the project interactively."""
        click.echo("Provide the values to generate the project documentation.")
        return super(DocsCMD, self).interactive_config()

    def record_files(self) -> None:
        """Register the files that will be generated by mloq."""
        source_files = {"conf.py", "index.md"}
        docs_ASSETS_path = Path("docs")
        for _file in self.files:
            if _file == deploy_docs:
                if not self.deploy_docs:
                    continue
                path = Path(".github") / "workflows"
            else:
                path = (
                    (docs_ASSETS_path / "source")
                    if str(_file.dst) in source_files
                    else docs_ASSETS_path
                )
            self.record.register_file(file=_file, path=path)



"""This module defines all the different assets accessible from mloq."""
from pathlib import Path
import sys
from typing import NamedTuple, Optional, Union


class File(NamedTuple):
    """
    Generates project files.

    This class defines templating files, which will be rendered according
    to the user's configuration. Besides, File instances have additional
    attributes used for specifying the destination of the generated file.

    Attributes of this class:
        name: Name of the templating file.
        src: Location of the templating file.
        dst: Name of the file generated from the templating file.
        description: Short description of the current file.
        is_static: Boolean value. If True, the templating file does not
            admit render parameters.
    """

    name: str
    src: Path
    dst: Path
    description: str
    is_static: bool


def file(
    name: str,
    path: Union[Path, str],
    description: Optional[str] = None,
    dst: Optional[Union[Path, str]] = None,
    is_static: bool = False,
) -> File:
    """Define a new asset as a File namedtuple."""
    if description is None:
        print("FIXME: %s must have a description" % name, file=sys.stderr)
        description = "TODO"
    dst = Path(dst) if dst is not None else name
    return File(
        name=name,
        src=Path(path) / name,
        dst=dst,
        description=description,
        is_static=is_static,
    )


# Assets paths
ASSETS_PATH = Path(__file__).parent / "assets"
SHARED_ASSETS_PATH = ASSETS_PATH / "shared"
MLOQ_ASSETS_PATH = ASSETS_PATH / "mloq"
REQUIREMENTS_PATH = ASSETS_PATH / "requirements"

# Mloq files
mloq_yml = file(
    "mloq.yaml",
    MLOQ_ASSETS_PATH,
    "mloq configuration, you can safely remove it if you don't plan to upgrade",
    is_static=True,
)
what_mloq_generated = file("WHAT_MLOQ_GENERATED.md", MLOQ_ASSETS_PATH, "this file")
# Requirements files

requirements = file(
    "requirements.txt",
    REQUIREMENTS_PATH,
    "list of exact versions of the packages on which your project depends",
    is_static=True,
)

dogfood_req = file(
    "dogfood.txt",
    REQUIREMENTS_PATH,
    "list of mock requirements for testing purposes",
    is_static=True,
)

# Shared templates
makefile = file("Makefile", SHARED_ASSETS_PATH, "common make commands for development")

pyproject_toml = file(
    "pyproject.toml",
    SHARED_ASSETS_PATH,
    "configuration of various development tools: linters, formatters, packaging",
)


def read_file(file: File) -> str:
    """Return and string with the content of the provided file."""
    with open(file.src, "r") as f:
        return f.read()



"""Mloq project command implementation."""
from pathlib import Path
from typing import Tuple

from mloq.command import Command
from mloq.config.param_patch import param
from mloq.files import ASSETS_PATH, file, makefile


PROJECT_ASSETS_PATH = ASSETS_PATH / "project"
readme = file("README.md", PROJECT_ASSETS_PATH, "README")
gitignore = file(
    ".gitignore",
    PROJECT_ASSETS_PATH,
    "list of files and directories ignored by Git operations",
    is_static=True,
)
pre_commit_hook = file(
    ".pre-commit-config.yaml",
    PROJECT_ASSETS_PATH,
    "Git pre-commit hooks configuration",
    is_static=True,
)
init = file(
    "init.txt",
    PROJECT_ASSETS_PATH,
    "Python package header",
    dst="__init__.py",
    is_static=True,
)
main = file(
    "main.txt",
    PROJECT_ASSETS_PATH,
    "Python package executable entry point",
    dst="__main__.py",
    is_static=True,
)
test_main = file(
    "test_main.txt",
    PROJECT_ASSETS_PATH,
    "Unit test of the python package executable entry point",
    dst="test_main.py",
    is_static=False,
)
test_req = file(
    "requirements-test.txt",
    PROJECT_ASSETS_PATH,
    "list of exact versions of the packages needed to run your test suite",
    is_static=True,
)
version = file(
    "version.txt",
    PROJECT_ASSETS_PATH,
    "defines the version of the package that is incremented on each push",
    dst="version.py",
    is_static=True,
)
code_of_conduct = file(
    "CODE_OF_CONDUCT.md",
    PROJECT_ASSETS_PATH,
    "behavioral rules and norms in open source projects",
)
codecov = file(
    ".codecov.yml",
    PROJECT_ASSETS_PATH,
    "configuration of CodeCov service to track the code coverage",
)
contributing = file(
    "CONTRIBUTING.md",
    PROJECT_ASSETS_PATH,
    "technical manual on how to contrib to the open source project",
)

PROJECT_FILES = [
    codecov,
    gitignore,
    readme,
    makefile,
    init,
    main,
    test_main,
    version,
    test_req,
    pre_commit_hook,
    contributing,
    code_of_conduct,
]

# TODO: select which files are created in config
# package folders: __init__, __main__, version
# contributing
# codecov
# gitignore
# makefile
# pre commit


class ProjectCMD(Command):
    """Implement the functionality of the project Command."""

    cmd_name = "project"
    files = tuple(PROJECT_FILES)
    disable = param.Boolean(default=False, doc="Disable project command?")
    project_name = param.String("${globals.project_name}", doc="Select project name")
    owner = param.String("${globals.owner}", doc="Github handle of the project owner")
    description = param.String("${globals.description}", doc="Short description of the project")
    project_url = param.String("${globals.project_url}", doc="GitHub project url")
    license = param.String("MIT", doc="Project license type")
    tests = param.Boolean(True, doc="Add support for pytest")

    @property
    def directories(self) -> Tuple[Path]:
        """Tuple containing Paths objects representing the directories created by the command."""
        project_folder = self.record.config.project.project_name.replace(" ", "_")

        return tuple([Path("src") / project_folder, Path("tests")])

    def record_files(self) -> None:
        """Register the files that will be generated by mloq."""
        self.record.register_file(file=readme, path=Path())
        self.record.register_file(file=makefile, path=Path())
        project_folder = Path("src") / self.record.config.project.project_name.replace(" ", "_")
        description = "Python package header for the project module"
        self.record.register_file(file=init, path=project_folder, description=description)
        self.record.register_file(file=main, path=project_folder)
        self.record.register_file(file=version, path=project_folder)
        description = "Python package header for the test module"
        self.record.register_file(file=init, path=Path("tests"), description=description)
        self.record.register_file(file=test_main, path=Path("tests"))
        root_files = [
            readme,
            makefile,
            test_req,
            pre_commit_hook,
            codecov,
            gitignore,
            code_of_conduct,
            contributing,
        ]
        for _file in root_files:
            self.record.register_file(file=_file, path=Path())



"""Mloq docker command implementation."""
from pathlib import Path
from typing import Optional

import click
from omegaconf import DictConfig, MISSING, OmegaConf

from mloq.command import Command
from mloq.commands.requirements import (  # REQUIREMENT_CHOICES,
    pytorch_req,
    RequirementsCMD,
    tensorflow_req,
)
from mloq.config.param_patch import param
from mloq.files import ASSETS_PATH, file


DOCKER_ASSETS_PATH = ASSETS_PATH / "docker"
dockerfile = file("Dockerfile", DOCKER_ASSETS_PATH, description="Docker container for the project")
makefile_docker = file(
    "Makefile.docker",
    DOCKER_ASSETS_PATH,
    description="Makefile for the Docker container setup",
    is_static=True,
)
DOCKER_FILES = [dockerfile, makefile_docker]


class DockerCMD(Command):
    """Implement the functionality of the docker Command."""

    cmd_name = "docker"
    files = tuple(DOCKER_FILES)
    disable = param.Boolean(default=None, doc="Disable docker command?")
    cuda = param.Boolean(None, doc="Install CUDA?")
    cuda_image_type = param.String(MISSING, doc="Type of cuda docker container")
    cuda_version = param.String("11.2", doc="CUDA version installed in the container")
    ubuntu_version = param.String("20.04", doc="Ubuntu version of the base image")
    project_name = param.String("${globals.project_name}", doc="Select project name")
    docker_org = param.String("${globals.owner}", doc="Name of your Docker organization")
    python_version = param.String("3.8", doc="Python version installed in the container")
    base_image = param.String(MISSING, doc="Base Docker image used to build the container")
    test = param.Boolean(True, doc="Install requirements-test.txt?")
    lint = param.Boolean(True, doc="Install requirements-lint.txt?")
    jupyter = param.Boolean(True, doc="Install a jupyter notebook server?")
    jupyter_password = param.String(
        "${docker.project_name}",
        doc="password for the Jupyter notebook server",
    )
    requirements = param.List(default=["none"], doc="Project requirements")
    extra = param.String("", doc="Extra code to add to Dockerfile")
    makefile = param.Boolean(True, doc="Add docker commands to makefile")
    # requirements = param.ListSelector(
    #    default="none", doc="Project requirements", objects=REQUIREMENT_CHOICES,
    # )

    @staticmethod
    def require_cuda_from_requirements(project_config: Optional[DictConfig] = None) -> bool:
        """Return True if any of the project dependencies require CUDA."""
        project_config = {} if project_config is None else project_config
        if "requirements" not in project_config:
            return False
        options = project_config.get("requirements", [])
        if RequirementsCMD.requirements_is_empty(options):
            return False
        elif isinstance(options, str):
            options = [options]
        tf_alias = RequirementsCMD.REQUIREMENTS_ALIASES[tensorflow_req]
        torch_alias = RequirementsCMD.REQUIREMENTS_ALIASES[pytorch_req]
        for option in options:
            if option in tf_alias or option in torch_alias:
                return True
        return False

    def requires_cuda(self) -> bool:
        """Return True if the Docker container requires CUDA."""
        if not OmegaConf.is_missing(self.config, "cuda") and self.cuda is not None:
            return self.config.cuda
        try:
            _ = self.config.requirements
        except Exception:
            self.config.requirements = []
        return self.require_cuda_from_requirements(self.config)

    def get_base_image(self):
        """Return the name of the base image for the project Docker container."""
        # Check value is not missing and it is not None due to a bad interpolation resolution
        if (
            not OmegaConf.is_missing(self.config, "base_image")
            and self.config.base_image is not None
        ):
            return self.config.base_image
        elif self.config.cuda:
            cuda_image = (
                f"nvidia/cuda:{self.config.cuda_version}-{self.config.cuda_image_type}"
                f"-ubuntu{self.config.ubuntu_version}"
            )
            return cuda_image
        return f"ubuntu:{self.config.ubuntu_version}"

    def parse_config(self) -> DictConfig:
        """Update the configuration dictionary from the data entered by the user."""
        super(DockerCMD, self).parse_config()
        if self.cuda is None:
            self.cuda = self.requires_cuda()
        self.base_image = self.get_base_image()
        return super(DockerCMD, self).parse_config()

    def interactive_config(self) -> DictConfig:
        """Generate the configuration of the project interactively."""
        click.echo("Provide the values to generate the project Docker container.")
        return self.parse_config()

    def record_files(self) -> None:
        """Register the files that will be generated by mloq."""
        self.record.register_file(file=dockerfile, path=Path())
        self.record.register_file(file=makefile_docker, path=Path())



"""Mloq package command implementation."""
from pathlib import Path

import click
from omegaconf import DictConfig, MISSING, OmegaConf

from mloq.command import Command
from mloq.config.param_patch import param
from mloq.files import ASSETS_PATH, file, pyproject_toml


PACKAGE_ASSETS_PATH = ASSETS_PATH / "package"
PYTHON_VERSIONS = ["3.6", "3.7", "3.8", "3.9", "3.10"]
DEFAULT_PYTHON_VERSIONS = ["3.7", "3.8", "3.9", "3.10"]
setup_py = file(
    "setup.txt",
    PACKAGE_ASSETS_PATH,
    "Python package installation metadata",
    dst="setup.py",
)
PACKAGE_FILES = [pyproject_toml, setup_py]


class PackageCMD(Command):
    """Implement the functionality of the package Command."""

    cmd_name = "package"
    files = tuple(PACKAGE_FILES)
    LICENSE_CLASSIFIERS = {
        "MIT": "License :: OSI Approved :: MIT License",
        "Apache-2.0": "License :: OSI Approved :: Apache Software License",
        "GPL-3.0": "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "proprietary": "License :: Other/Proprietary License",
    }
    disable = param.Boolean(default=None, doc="Disable package command?")
    pyproject_extra = param.String("", doc="Additional pyproject.toml configuration")
    project_name = param.String("${globals.project_name}", doc="Select project name")
    license = param.String("MIT", doc="Project license type")
    license_classifier = param.String(MISSING, doc="License classifier in setup.py")
    description = param.String("${globals.description}", doc="Short description of the project")
    default_branch = param.String(doc="Default branch of the project")
    project_url = param.String("${globals.project_url}", doc="GitHub project url")
    owner = param.String("${ci.author}", doc="Github handle of the project owner")
    author = param.String(doc="Author(s) of the project")
    email = param.String(doc="Owner contact email")
    python_versions = param.ListSelector(
        default=DEFAULT_PYTHON_VERSIONS,
        doc="Supported python versions",
        objects=PYTHON_VERSIONS,
    )
    use_poetry = param.Boolean(
        "${globals.use_poetry}",
        doc="Use poetry to manage dependencies",
    )

    def parse_config(self) -> DictConfig:
        """Update the configuration DictConfig with the Command parameters."""
        conf = super(PackageCMD, self).parse_config()
        if OmegaConf.is_missing(conf.package, "license_classifier"):
            conf.package.license_classifier = self.LICENSE_CLASSIFIERS.get(
                conf.package.get("license", "proprietary"),
                "",
            )
        return conf

    def interactive_config(self) -> DictConfig:
        """Generate the configuration of the project interactively."""
        click.echo("Provide the values to generate the packaging files.")
        return self.parse_config()

    def record_files(self) -> None:
        """Register the files that will be generated by mloq."""
        for _file in self.files:
            self.record.register_file(file=_file, path=Path())



"""Package for initializing ML projects following ML Ops best practices."""
from datetime import date
import logging

import flogging
from omegaconf import OmegaConf


flogging.setup()
_logger = logging.getLogger("mloq")
OmegaConf.register_new_resolver("current_year", lambda: date.today().year)



"""This module contains some utilities that are not currently used."""
import filecmp
import os
from pathlib import Path
from typing import List, Union

from omegaconf import DictConfig, OmegaConf

from mloq.files import mloq_yml


# TODO: Fix docs
def dir_trees_are_equal(dir1: Union[str, Path], dir2: Union[str, Path]) -> bool:
    """
    Compare two directories recursively. Files in each directory are \
    assumed to be equal if their names and contents are equal.

    @param dir1: First directory path
    @param dir2: Second directory path

    @return: True if the directory trees are the same and
        there were no errors while accessing the directories or files,
        False otherwise.
    """
    dirs_cmp = filecmp.dircmp(dir1, dir2)
    if (
        len(dirs_cmp.left_only) > 0
        or len(dirs_cmp.right_only) > 0
        or len(dirs_cmp.funny_files) > 0
    ):
        return False
    (_, mismatch, errors) = filecmp.cmpfiles(dir1, dir2, dirs_cmp.common_files, shallow=False)
    if len(mismatch) > 0 or len(errors) > 0:
        return False
    for common_dir in dirs_cmp.common_dirs:
        new_dir1 = os.path.join(dir1, common_dir)
        new_dir2 = os.path.join(dir2, common_dir)
        if not dir_trees_are_equal(new_dir1, new_dir2):
            return False
    return True


def files_are_equal(path1: Union[str, Path], path2: Union[str, Path]) -> bool:
    """
    Compare the content of two files.

    Compare two incoming files. They are assumed equal if their contents
    are the same.

    Args:
          path1: Path containing the first file to be compared.
          path2: Path containing the second file to be compared.

    Return:
         It returns True if the two given files are equal and no errors
            have arisen during the process. False otherwise.
    """
    path1 = Path(path1) if isinstance(path1, str) else path1
    path2 = Path(path2) if isinstance(path2, str) else path2
    return filecmp.cmp(path1, path2, shallow=False)


def get_generated_files(path: Union[str, Path]) -> List[Path]:
    """
    List all the files generated in the last mloq run.

    Args:
        path: path to WHAT_MLOQ_GENERATED.md file.

    Returns:
        List of Path containing the names of the files generated by mloq.
    """
    # Read file as string
    # Split file to end up with the desired file names
    pass


# TODO: Improve the Ledger class so it can also track the different directories
#  that were created, and add them under a new section in the WHAT_MLOQ_GENERATED.md


def get_generated_directories(path: Union[str, Path]) -> List[Path]:
    """
    List all the directories generated in the last mloq run.

    Args:
        path: path to WHAT_MLOQ_GENERATED.md file.

    Returns:
        List of Path containing the names of the directories generated by mloq.
    """
    pass


def check_directories_exist(paths: List[Union[str, Path]]) -> bool:
    """
    Check if the provided paths exist.

    Args:
        paths: List of paths that will be checked

    Returns:
        True if all the provided paths exist. False otherwise.
    """
    for path in paths:
        path = Path(path) if isinstance(path, str) else path
        cond1 = os.path.exists(path)
        cond2 = os.path.isfile(path) or os.path.isdir(path)
        cond3 = Path.exists(path)
        if not (cond1 and cond2 and cond3):
            print(f"The given path {path} is neither a file nor a directory.")
        return cond1 and cond2 and cond3


def get_docker_python_version(template: DictConfig) -> str:
    """Return the highest python version defined for the project."""
    max_version = list(sorted(template["python_versions"]))[-1]
    version = max_version.replace(".", "")
    return f"py{version}"


def write_config_setup(config: DictConfig, path: Union[Path, str], safe: bool = False):
    """Write setup config in a yaml file."""
    if safe:
        path = Path(path)
        path = path / mloq_yml.dst if path.is_dir() else path
    with open(path, "w") as f:
        OmegaConf.save(config, f)


def load_empty_config_setup() -> DictConfig:
    """Return a dictionary containing all the MLOQ setup config values set to None."""
    return OmegaConf.load(mloq_yml.src)



"""Mloq ci command implementation."""
from pathlib import Path
from typing import Tuple

import click
from omegaconf import DictConfig
import param

from mloq.command import Command
from mloq.commands.package import DEFAULT_PYTHON_VERSIONS, PYTHON_VERSIONS
from mloq.files import ASSETS_PATH, file


CI_ASSETS_PATH = ASSETS_PATH / "ci"
push_python_wkf = file(
    "push.yml",
    CI_ASSETS_PATH,
    "GitHub Actions continuous integration workflow file",
)

WORKFLOW_FILES = [push_python_wkf]


class CiCMD(Command):
    """Implement the functionality of the ci Command."""

    cmd_name: str = "ci"
    ubuntu_version = param.String(doc="Primary Ubuntu version in GitHub Actions")
    disable = param.Boolean(default=False, doc="Disable ci command?")
    docker = param.Boolean(doc="Test Docker container in the CI?")
    project_name = param.String("${globals.project_name}", doc="Select project name")
    default_branch = param.String(doc="Default branch of the project")
    docker_org = param.String(doc="Name of your docker organization")
    bot_name = param.String(doc="Bot's GitHub login to push commits in CI")
    bot_email = param.String(doc="Bot account email")
    ci_python_version = param.String(doc="Primary Python version in GitHub Actions")
    python_versions = param.ListSelector(
        default=DEFAULT_PYTHON_VERSIONS,
        doc="Supported python versions",
        objects=PYTHON_VERSIONS,
    )
    ci_extra = param.String(
        doc="Additional script in GitHub Actions before running the main tests",
    )
    vendor = param.String(doc="Continuous Integration Vendor")
    open_source = param.Boolean(doc="Is the project Open Source?")
    author = param.String(doc="Author(s) of the project")
    owner = param.String("${ci.author}", doc="Github handle of the project owner")
    email = param.String(doc="Owner contact email")
    project_url = param.String(doc="GitHub project url")
    files = tuple(WORKFLOW_FILES)

    @property
    def directories(self) -> Tuple[Path]:
        """Tuple containing Paths objects representing the directories created by the command."""
        return tuple([Path(".github") / "workflows"])

    def interactive_config(self) -> DictConfig:
        """Generate the configuration of the project interactively."""
        click.echo("Provide the values to set up continuous integration.")
        return super(CiCMD, self).interactive_config()

    def record_files(self) -> None:
        """Register the files that will be generated by mloq."""
        self.record.register_file(file=push_python_wkf, path=Path(".github") / "workflows")



"""Mloq requirements command implementation."""
from pathlib import Path
import tempfile
from typing import Iterable, List, Union

import click
from omegaconf import DictConfig

from mloq.command import Command
from mloq.config.param_patch import param
from mloq.files import ASSETS_PATH, File, file
from mloq.record import CMDRecord


# Requirements files
REQUIREMENTS_PATH = ASSETS_PATH / "requirements"
requirements = file(
    "requirements.txt",
    REQUIREMENTS_PATH,
    "list of exact versions of the packages on which your project depends",
    is_static=True,
)
data_science_req = file(
    "data-science.txt",
    REQUIREMENTS_PATH,
    "list of commonly used data science libraries",
)
data_viz_req = file(
    "data-visualization.txt",
    REQUIREMENTS_PATH,
    "list of commonly used visualization libraries",
    is_static=True,
)
pytorch_req = file(
    "pytorch.txt",
    REQUIREMENTS_PATH,
    "Pytorch deep learning libraries",
    is_static=True,
)
tensorflow_req = file(
    "tensorflow.txt",
    REQUIREMENTS_PATH,
    "Tensorflow deep learning libraries",
    is_static=True,
)
lint_req = file(
    "requirements-lint.txt",
    REQUIREMENTS_PATH,
    "list of exact versions of the packages used to check your code style",
    is_static=True,
)
test_req = file(
    "requirements-test.txt",
    REQUIREMENTS_PATH,
    "list of exact versions of the packages needed to run your test suite",
    is_static=True,
)
dogfood_req = file(
    "dogfood.txt",
    REQUIREMENTS_PATH,
    "list of mock requirements for testing purposes",
    is_static=True,
)
docs_req = file(
    "requirements-docs.txt",
    REQUIREMENTS_PATH,
    "list of exact versions of the packages needed to build your documentation",
    is_static=True,
)
REQUIREMENTS_FILES = [pytorch_req, data_science_req, data_viz_req, tensorflow_req]

REQUIREMENT_CHOICES = [
    "data-science",
    "data-viz",
    "torch",
    "tensorflow",
    "none",
    "dogfood",
    "None",
]


class RequirementsCMD(Command):
    """Implement the functionality of the requirements Command."""

    cmd_name = "requirements"
    files = tuple(REQUIREMENTS_FILES)
    disable = param.Boolean(default=None, doc="Disable requirements command?")
    requirements = param.ListSelector(
        default=["none"],
        doc="Project requirements",
        objects=REQUIREMENT_CHOICES,
    )
    REQUIREMENTS_ALIASES = {
        data_science_req: ["data-science", "datascience", "ds"],
        pytorch_req: ["pytorch", "torch"],
        tensorflow_req: ["tensorflow", "tf"],
        data_viz_req: ["data-visualization", "data-viz", "data-vis", "dataviz", "datavis"],
    }

    def __init__(self, record: CMDRecord, interactive: bool = False):
        """
        Initialize a RequirementsCMD class.

        Args:
            record: CMDRecord where the command data will be written.
            interactive: If True, parse the command configuration in interactive mode.
        """
        super(RequirementsCMD, self).__init__(record=record, interactive=interactive)
        self._temp_dir = tempfile.TemporaryDirectory()
        # File objects work referencing files present in the system. We create a requirements.txt
        # temporary file to be consistent with that behavior.
        reqs_src = Path(self._temp_dir.name) / "requirements.txt"
        self._reqs_file = File(
            name=requirements.name,
            src=reqs_src,
            dst=requirements.dst,
            is_static=requirements.is_static,
            description=requirements.description,
        )
        self.files = tuple(list(self.files) + [self._reqs_file])

    def __del__(self) -> None:
        """Remove the temporary directory when the instance is deleted."""
        self._temp_dir.cleanup()

    @classmethod
    def get_aliased_requirements_file(cls, option: str) -> File:
        """Get requirement file from aliased name."""
        for _file, valid_alias in cls.REQUIREMENTS_ALIASES.items():
            if option in valid_alias:
                return _file
        if option == "dogfood":
            return dogfood_req
        raise KeyError(
            f"{option} is not a valid name. Valid aliases are {cls.REQUIREMENTS_ALIASES}",
        )

    @classmethod
    def read_requirements_file(cls, option: str) -> str:
        """Return the content of the target requirements file form an aliased name."""
        req_file = cls.get_aliased_requirements_file(option)
        with open(req_file.src, "r") as f:
            return f.read()

    @classmethod
    def compose_requirements(cls, options: Iterable[str]) -> str:
        """
        Return the content requirements.txt file with pinned dependencies.

        The returned string contains the combined dependencies\
         for the different options sorted alphabetically.

        Args:
            options: Iterable containing the aliased names of the target \
                     dependencies for the project.

        Returns:
            str containing the pinned versions of all the selected requirements.
        """
        requirements_text = ""
        for i, opt in enumerate(options):
            pref = "\n" if i > 0 else ""  # Ensure one requirement per line
            requirements_text += pref + cls.read_requirements_file(opt)
        # Sort requirements alphabetically
        requirements_text = "\n".join(sorted(requirements_text.split("\n"))).lstrip("\n")
        return requirements_text

    '''def ___parse_config(self) -> DictConfig:
        """Update the configuration DictConfig with the Command parameters."""
        value = self.CONFIG.requirements(self.record.config, self.interactive)
        self.record.config.requirements = value
        return self.record.config'''

    def interactive_config(self) -> DictConfig:
        """Generate the configuration of the project interactively."""
        click.echo("Please specify the requirements of the project as a comma separated list.")
        click.echo("Available values:")
        click.echo(
            "    data-science: Common data science libraries such as numpy, pandas, sklearn...",
        )
        click.echo(
            (
                "    data-viz: Visualization libraries such as holoviews, ",
                "bokeh, plotly, matplotlib...",
            ),
        )
        click.echo("    pytorch: Latest version of pytorch, torchvision and pytorch_lightning")
        click.echo("    tensorflow: ")  # , data-viz, torch, tensorflow}")
        return self.parse_config()

    @staticmethod
    def requirements_is_empty(options: Union[List[str], str]) -> bool:
        """Return True if no requirements are specified for the project."""
        if not options:
            return True
        if isinstance(options, str):
            options = [options]
        if None in options or "None" in options or "none" in options:
            return True
        return False

    def record_files(self) -> None:
        """Register the files that will be generated by mloq."""
        reqs_value = self.record.config.requirements.requirements
        if self.requirements_is_empty(reqs_value):
            return
        reqs_content = self.compose_requirements(reqs_value)
        with open(self._reqs_file.src, "w") as f:
            f.write(reqs_content)

        self.record.register_file(file=self._reqs_file, path=Path())



"""This module defines the base Command class used for defining mloq commands."""
from pathlib import Path
from typing import Tuple

from omegaconf import DictConfig

from mloq.config.configuration import as_resolved_dict
from mloq.config.prompt import Promptable
from mloq.writer import CMDRecord


class CommandMixin:
    """Class containing the interface for defining an MLOQ Command."""

    files: tuple = tuple()
    cmd_name = "command"

    def __init__(self, record: CMDRecord, interactive: bool = False, *args, **kwargs):
        """
        Instantiate the Command class.

        Args:
            record: CMDRecord instance. Keeps a record of the user's
                configuration. It registers the files and directories
                that will be generated by mloq.
            interactive: Boolean value. If True, configuration values are
                introduced interactively. Otherwise, the configuration is
                parsed from a user-generated configuration file.
        """
        self._record = record
        self.interactive = interactive
        super().__init__(*args, **kwargs)

    @property
    def record(self) -> CMDRecord:
        """Return a CMDRecord that keeps track of the files and directories the command creates."""
        return self._record

    @property
    def directories(self) -> Tuple[Path]:
        """
        Tuple containing Paths objects representing the directories the Command creates.

        Override this property if your command creates any directories.

        Returns:
            Tuple of :class:`Path` objects representing the path to the directories that the
            :class:`Command` will create.
        """
        return tuple()

    def parse_config(self) -> DictConfig:
        """
        Update the configuration dictionary from the data entered by the user.

        Given the basic configuration skeleton (contained in mloq.yaml), \
        this method updates the values of those parameters (included in \
        CONFIG object) that are related to the selected command. Incoming \
        values are introduced either interactively or via a custom user's \
        mloq.yaml file.

        Returns:
            It returns an updated version of the 'config' attribute of \
            the 'record' instance.
        """
        self.record.update_config(DictConfig({self.cmd_name: self.config}))
        return self.record.config

    def interactive_config(self) -> DictConfig:
        """Pass user's configuration interactively."""
        raise NotImplementedError

    def record_files(self) -> None:
        """Register the files that will be generated by mloq."""
        if len(self.files) > 0:
            raise NotImplementedError

    def record_directories(self) -> None:
        """Register the directories that will be generated by mloq."""
        for directory in self.directories:
            self._record.register_directory(directory)

    def configure(self) -> None:
        """
        Save the updated version of the 'config' attribute.

        After parsing the new configuration values introduced by the user,
        this method registers and saves this updated configuration within
        the '_config' attribute of the 'record' instance.
        """
        if self.interactive:
            self.interactive_config()
            # config = self.interactive_config()
        else:
            self.parse_config()
            # config = self.parse_config()
        # self.record.update_config(config)

    def run_side_effects(self) -> None:
        """Apply additional configuration methods."""
        pass

    def run(self) -> CMDRecord:
        """
        Record the files and directories generated by mloq according to the user's configuration.

        This method updates the configuration dictionary with the values
        introduced by the user. Once the parameters have been revised,
        the files and directories that will be generated by mloq are
        registered within the 'record' instance.

        Returns:
            It returns an updated version of the CMDRecord instance, where
                the files and directories that will be generated by mloq
                are recorded within the 'record' instance.
        """
        self.configure()
        self.record_directories()
        self.record_files()
        self.run_side_effects()
        return self.record


class Command(CommandMixin, Promptable):
    """
    Define blueprints for generating custom mloq commands.

    Base class used for defining new mloq commands. It establishes the
    fundamental methods for defining and updating the configuration values
    used to create the necessary files for the user's project, while
    registering the latter for later use.

    This class is initialized from a CMDRecord instance, object where the
    user's configuration, as well as the files and directories that will
    be generated, are stored.

    class Attributes:
        name: Name of the command.
        files: Tuple containing the templates used for creating the \
               necessary files of your project.
        CONFIG: NamedTuple containing the keys and values of your \
                configuration RELATIVE to the command.
    """

    def __init__(self, record: CMDRecord, interactive: bool = False, **kwargs):
        """
        Instantiate the Command class.

        Args:
            record: CMDRecord instance. Keeps a record of the user's
                configuration. It registers the files and directories
                that will be generated by mloq.
            interactive: Boolean value. If True, configuration values are
                introduced interactively. Otherwise, the configuration is
                parsed from a user-generated configuration file.
            cfg_node: key of the DictConfig in record that contains the configuration
                      of the current command.
        """
        # self._record = record
        # self.interactive = interactive
        # TODO: allow interpolations in config that refer to the global record config
        # cmd_conf = self._config_from_record()
        # cfg_node = self.cmd_name if cfg_node is None else cfg_node
        super(Command, self).__init__(
            record=record,
            interactive=interactive,
            config=record.config,
            cfg_node=self.cmd_name,
            **kwargs,
        )

    def interactive_config(self) -> DictConfig:
        """Pass user's configuration interactively."""
        prompt_conf = DictConfig({self.cmd_name: self.prompt.prompt_all()})
        self.record.update_config(prompt_conf)
        return self.record.config

    def _config_from_record(self) -> DictConfig:
        return DictConfig(as_resolved_dict(self.record.config))[self.cmd_name]
